package main;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.nio.file.Path;

import javax.imageio.ImageIO;

public class EdgeDetection {
	
	private static Image applyFilter(Image image, float[][] kernel, boolean linearFilter) {
		float[][] imageData = image.getData();
		Image clonedImage = image.clone();
		
		double multiplikator = 1;
		if (linearFilter == true) {
			multiplikator = 1.0/16.0;
		}
		
		for (int BildY = 0; BildY < imageData.length; BildY++) {
			for (int BildX = 0; BildX < imageData[BildY].length; BildX++) {
				float sum = 0;
				
				for (int KernelX = 0; KernelX < kernel.length; KernelX++) {
					for (int KernelY = 0; KernelY < kernel[KernelX].length; KernelY++) {
						int middleX = Math.round((kernel.length - 1) / 2);
						int middleY = Math.round((kernel[KernelX].length -1) / 2);
						
						int positionX = BildX + KernelY - middleY;
						int positionY = BildY + KernelX - middleX;
						
						if (positionX < 0) {
							positionX = 0;
						}
						if (positionY >= imageData.length) {
							positionY = imageData.length - 1;
						}
						if (positionY < 0) {
							positionY = 0;
						}
						if (positionX >= imageData[BildY].length) {
							positionX = imageData[BildY].length - 1;
						}
						
						float id = imageData[positionY][positionX];
						float k = kernel[KernelX][KernelY];
						sum +=  id * (multiplikator) * k;
					}
				}
				
				int sumAsInteger = (int) sum;
				sum = sumAsInteger;
				
				if (sum < 0) {
					sum = 0;
				}
				
				if (sum > image.getMax()) {
					sum = image.getMax();
				}
				
				clonedImage.setPixel(sum, BildY, BildX);
				//image.setPixel(sum, BildY, BildX);
			}
		}
		
		return clonedImage;
	}

	public static Image linearFilter(Image image, float[][] kernel) {
		// TODO: H1
		Image clonedImage = applyFilter(image, kernel, true);
		
		return clonedImage;
	}

	public static Image sobel(Image input, boolean x) {
		// TODO: H2
		float[][] kernelX = {
				{-1, 0, 1},
				{-2, 0, 2},
				{-1, 0, 1},
		};
		
		float[][] kernelY = {
				{1, 2, 1},
				{0, 0, 0},
				{-1, -2, -1},
		};
		
		float[][] kernel;
		
		// Kernel fr x-Achse
		if (x == true) {
			kernel = kernelX;
		} else {
			kernel = kernelY;
		}
		
		Image clonedImage = applyFilter(input, kernel, false);
		
		return clonedImage;
	}

	public static Image edgeDirection(Image sobelX, Image sobelY) {
		// TODO: H3
		float[][] imageDataX = sobelX.getData();
		float[][] imageDataY = sobelY.getData();
		Image newImage = sobelX.clone();
		
		for (int y = 0; y < imageDataX.length; y++) {
			for (int x = 0; x < imageDataX[y].length; x++) {
				float gX = imageDataX[y][x];
				float gY = imageDataY[y][x];
				
				float theta = (float) Math.toDegrees(Math.atan2(gY, gX));

				if ((theta >= 0 && theta <= 22.5) || 
						(theta >= 157.5 && theta <= 180)) {
					theta = 0;
				}
				
				if (theta >= 22.5 && theta <= 67.5) {
					theta = 45;
				}
				
				if (theta >= 67.5 && theta <= 112.5) {
					theta = 90;
				}
				
				if (theta >= 112.5 && theta >= 157.5) {
					theta = 135;
				}
				
				newImage.setPixel(theta, y, x);
			}
		}
		return newImage;
	}

	public static Image gradient(Image sobelX, Image sobelY) {
		// TODO: H4
		float[][] imageDataX = sobelX.getData();
		float[][] imageDataY = sobelY.getData();
		Image newImage = sobelX.clone();
		
		for (int y = 0; y < imageDataX.length; y++) {
			for (int x = 0; x < imageDataX[y].length; x++) {
				float gX = imageDataX[y][x];
				float gY = imageDataY[y][x];
				
				float gCalc = (float) Math.sqrt((gX * gX) + (gY * gY));
				int gCalcAsInt = (int) gCalc;
				gCalc = gCalcAsInt;
				
				newImage.setPixel(gCalc, y, x);
			}
		}
		return newImage;
	}

	public static Image nonMaximumSupression(Image gradient, Image theta) {
		float[][] imageDataGradient = gradient.getData();
		float[][] imageDataTheta = theta.getData();
		Image newImage = gradient.clone();
		
		for (int y = 0; y < imageDataGradient.length; y++) {
			for (int x = 0; x < imageDataGradient[y].length; x++) {
				float g = imageDataGradient[y][x];
				
				// Wenn Rand, dann Pixel = 0
				if (y == 0) {
					// Y = 0
					newImage.setPixel(0, y, x);
				} else if (x == 0) {
					// X = 0
					newImage.setPixel(0, y, x);
				} else if (y == (imageDataGradient.length - 1)) {
					// Y am Ende
					newImage.setPixel(0, y, x);
				} else if (x == (imageDataGradient[y].length - 1)) {
					// X am Ende
					newImage.setPixel(0, y, x);
				} else {
					// Hole Nachbarn Pixel
					float neighbor1 = 0;
					float neighbor2 = 0;
					
					if (imageDataTheta[y][x] == 0) {
						// stlich - Westlich
						neighbor1 = imageDataGradient[y][x+1]; // stlich
						neighbor2 = imageDataGradient[y][x-1]; // Westlich
					} else if (imageDataTheta[y][x] == 90) {
						// Nrdlich - Sdlich
						neighbor1 = imageDataGradient[y-1][x]; // Nrdlich
						neighbor2 = imageDataGradient[y+1][x]; // Sdlich
					} else if (imageDataTheta[y][x] == 135) {
						// Nordwestlich - Sdstlich
						neighbor1 = imageDataGradient[y-1][x-1]; // Nordwestlich
						neighbor2 = imageDataGradient[y+1][x+1]; // Sdstlich
					} else if (imageDataTheta[y][x] == 45) {
						// Nordstlich - Sdwestlich
						neighbor1 = imageDataGradient[y-1][x+1]; // Nordstlich
						neighbor2 = imageDataGradient[y+1][x-1]; // Sdwestlich
					}
					
					if (neighbor1 > g || neighbor2 > g) {
						newImage.setPixel(0, y, x);
					}
				}
				
			}
		}
		return newImage;
	}

	public static Image hysteresis(Image gradient, float t1, float t2) {
		// TODO: H6s
		float[][] imageDataGradient = gradient.getData();
		Image newImage = gradient.clone();
		
		for (int y = 1; y < (imageDataGradient.length - 1); y++) {
			for (int x = 1; x < (imageDataGradient[y].length - 1); x++) {
				float g = imageDataGradient[y][x];
				
				if (g <= t1) {
					// Hole Nachbarn Pixel
					float neighborNW = imageDataGradient[y-1][x-1];
					float neighborN =  imageDataGradient[y-1][x];
					float neighborNO = imageDataGradient[y-1][x+1];
					float neighborO =  imageDataGradient[y][x+1];
					float neighborSO = imageDataGradient[y+1][x+1];
					float neighborS =  imageDataGradient[y+1][x];
					float neighborSW = imageDataGradient[y+1][x-1];
					float neighborW =  imageDataGradient[y][x-1];
					
					if (neighborNW < t2 && neighborN < t2 && neighborNO < t2 &&
							neighborO < t2 &&
							neighborSO < t2 && neighborS < t2 && neighborSW < t2 &&
							neighborW < t2) {
						// Alle Pixel sind kleiner t2 (Keins ist >= t2)
						// Setze den aktuellen Pixel auf 0
						newImage.setPixel(0, y, x);
					}
				}
			}
		}
		return newImage;
	}

	public static Image detectEdges(Image input, float t1, float t2) {
		// TODO: H7.1
		// 1)
		float[][] linearKernel = {
				{1, 2, 1},
				{2, 4, 2},
				{1, 2, 1}
		};
		Image imageWithLinearFilter = linearFilter(input, linearKernel);
		
		// 2)
		Image imageSobelX = sobel(imageWithLinearFilter, true);
		Image imageSobelY = sobel(imageWithLinearFilter, false);

		// 3)
		Image imageTheta = edgeDirection(imageSobelX, imageSobelY);
		
		// 4)
		Image imageGradient = gradient(imageSobelX, imageSobelY);

		// 5)
		Image imageNonMaximum = nonMaximumSupression(imageGradient, imageTheta);
		
		// 6)
		Image imageHysteresis = hysteresis(imageNonMaximum, 130, 200);
		
		return imageHysteresis;
	}

	public static void saveEdgeOverlay(Image input, float t1, float t2, Path path) throws IOException {
		// TODO: H7.2
		Image detectedEdges = detectEdges(input, t1, t2);
		BufferedImage rgbImage = new BufferedImage(input.getWidth(), input.getHeight(), BufferedImage.TYPE_INT_ARGB);
		
		float[][] inputData = input.getData();
		float[][] detectedEdgesData = detectedEdges.getData();
		
		for (int y = 0; y < inputData.length; y++) {
			for (int x = 0; x < inputData[y].length; x++) {
				float inputDataPixel = inputData[y][x];
				float edge = detectedEdgesData[y][x];
				
				if (edge > 0) {
					rgbImage.setRGB(x, y, Color.MAGENTA.getRGB());
				} else {
					Color c = new Color(inputDataPixel / 255.0f, inputDataPixel / 255.0f, inputDataPixel / 255.0f, 1);
					rgbImage.setRGB(x, y, c.getRGB());
				}
			}
		}
		
		ImageIO.write(rgbImage, "PNG", path.toFile());
	}

}
